

function errorf(str,...)
  return error(string.format(str,...),2)
end

function errorf2(str,...)
  return error(string.format(str,...),3)
end

function errorf3(str,...)
  return error(string.format(str,...),4)
end

local Log, log0, log in gui.BattleLog
local is_object, systems,weapons,bernoulli_trials,random_chance,queue_death,isnumber,istable,drawers in _ENV
local min,max,ceil in math

depends '@legacy_iterators'

local function min_hull_integrity(target)
  return floor(target.structure*.3)
end

local function get_hull(target)
  return target.structure-(target.structure_damage or 0)
end


function heal_capture_damage(ship)
  for s in systems(ship)
    if s.pre_capture_damage
      s.damage = s.pre_capture_damage or s.damage
    end
  end

  for s in weapons(ship)
    if s.pre_capture_damage
      s.damage=s.pre_capture_damage or s.damage
    end
  end
end




function post_capture_damage(ship)

  if ship.empire:known 'Quantum Decryption'
    return
  end

  for s in systems(ship)
    s.pre_capture_damage=s.damage
    s.damage=s.health
  end

  for s in weapons(ship)
    s.pre_capture_damage=s.damage
    --s.damage=s.health_per*s.n   
    s.damage=s.health
  end

  ship.cloaked=nil

  update_ship_speed(ship)
  update_ship_shields(ship)
end

function update_ship_speed(ship)

  local drives, total_drives =0,0

  for s in systems(ship)
   
    if s.drive
      --print(s,'is a drive',s.health,s.damage)

      total_drives+=s.health
      drives+=s.health-s.damage

    end
  end

  --print('drive health',drives/total_drives)
  if not ship.max_speed_class
    ship.max_speed_class = ship.speed_class
  end

  if ship.max_speed_class > 0
    ship.speed_class = round | ship.max_speed_class * drives / total_drives
  end
  ship.fuel = min(ship.fuel,ship.max_fuel)
  ship.motion_speed = ship.max_fuel

  -- if we've been immobilized, we can't retreat.
	if ship.speed_class==0
		ship.retreating=nil
	end

end

function ShipProperty.shield_regen_rate(ship)
  local regenerators=0

  for s in systems(ship)
    if (s.type == 'shield_regenerator') and (s.damage < .5*s.health)
      regenerators++
    end
  end

  -- more or less fair?
  return .25 + .75 * ( 1.0 - .5^(regenerators) )
end

function ShipProperty.adjusted_max_shields(ship)
  local max_shields in ship
  if max_shields==0
    return 0
  end
  local shield_health=0
  local shield_damage=0

  for s in systems(ship)
    if s.is_shield
      shield_health+=s.health
      shield_damage+=s.damage
    end
  end
  
  return round(max_shields - max_shields * shield_damage / shield_health)
end

function shield_regen(ship)

  local adjusted_max_shields in ship

  if adjusted_max_shields == 0
    return
  end

  local regen=floor(adjusted_max_shields*ship.shield_regen_rate)

--	print(ship.name,ship.shield_regen_rate,regen,adjusted_max_shields)


  local needs_ani

  for i,v in ipairs(SHIELD_NAMES) do
    needs_ani=ship[v]~=0 or needs_ani
    ship[v]=max(0,ship[v]-regen)
  end

  return needs_ani
end

--print(..objects['Oldish Cruiser'].capacitor_charge)

function ShipProperty.has_functioning_autorepair(ship)
  local autorepair_systems = 0
  for s in systems(ship)
    if s.repair_rate and ((s.damage or 0) < s.health)
			return true
    end
  end
end

function ShipProperty.autorepair_missiles(ship)
	return drawers.safe_get_theme(ship.drawer)=='Hat'
end

function ShipProperty.autorepair_fighters(ship)
	return drawers.safe_get_theme(ship.drawer)=='Hat'
end

function apply_autorepair(ship)
  local autorepair_systems = ship.drawer_table?.autorepair_rate or 0
  for s in systems(ship)
    if s.repair_rate
      autorepair_systems += (1 - s.damage/s.health) * s.repair_rate
    end
  end

  if autorepair_systems==0
    return
  end

  local heal = min(1, autorepair_systems)

  for s in systems(ship)
    local damage,health in s
    s.damage = max(0, round | damage-health*heal)
  end

  for s in weapons(ship)
    local damage,health,health_per in s
    needs_ani = n
    s.damage = max(0, health_per * round | (damage-health*heal) / health_per  )
  end


  local structure_damage, armor_damage in ship

  ship.structure_damage = max(0, round | ship.structure_damage - ship.structure*heal)
  ship.armor_damage = max(0, round | ship.armor_damage - ship.armor*heal)

	--!! trigger missile/fighter autorepairs if appropriate
	if (ship.autorepair_missiles) and ship.munitions
		local max_munitions in ship
		ship.munitions=min( ship.munitions + round(max_munitions*heal), max_munitions)
	end

	---[[
	if ship.autorepair_fighters
		local remainder = 0
		for launch_bay in weapons(ship)
			if launch_bay.type=='fighter'
				local n,missing in launch_bay
				launch_bay.regen_missing =  n*heal + (launch_bay.regen_missing or 0)
				if launch_bay.regen_missing > missing
					launch_bay.regen_missing =missing
				end
	
			if launch_bay.regen_missing>=1
					local heal = floor(launch_bay.regen_missing)
					launch_bay.missing -= heal
					launch_bay.regen_missing -= heal
				end

				print('MISSING',missing,launch_bay.missing)
				--error 'thats the case'
			end
		end
	end
	--]]

  local needs_ani = (structure_damage~=ship.structure_damage) or (armor_damage~=ship.armor_damage) 
  --print(..needs_ani)
  return needs_ani
end


function update_ship_shields(ship)
  local adjusted_max_shields,max_shields in ship
  local min_damage = max_shields-adjusted_max_shields
  for i,v in ipairs(SHIELD_NAMES)
    ship[v]=max( ship[v]  , min_damage)
  end
end


function heal_emp(ship)
  local rval
  for s in systems(ship) do

    rval=rval or s.emp_damage

    if(s.emp_damage)
      s.damage -= s.emp_damage
      if(s.damage<0) s.damage=0 end
      s.emp_damage = nil

      local sinfo = SYSTEMS[s.parent_type]

      -- recover speed if the ship's drive was damaged by emp
      --if(sinfo.drive and ship.undamaged_speed)
      --  ship.speed_class=ship.undamaged_speed 
      --end

    end

  end

  update_ship_speed(ship)

  for w in weapons(ship) do

    rval=rval or w.emp_damage

    if(w.emp_damage)
      w.damage -= w.emp_damage
      if(w.damage<0) w.damage=0 end
      w.emp_damage = nil
    end

  end

  return rval

end

function can_self_destruct(ship)

  if(ship.dead or ship.transport) return false end

  if ship.is_planet return false end

  for s in systems(ship)
    if (s.type=='reactor') and (s.health > s.damage)
      return true
    end
  end

  return false

end

function is_max_crippled(w,target)
  if(is_ship(target))

    if w:special "EMP"

      --sprint(target.structure-target.structure_damage-1,.2*target.structure )

      if(target.dead) return true end

      if(get_hull(target)-1 < min_hull_integrity(target) )
        return current_system_and_weapon_health(target)==0
      end

      return false

    else
      return target.dead or target.critical_condition or ( get_hull(target)-1 < min_hull_integrity(target) ) 
    end

  else
    return target.dead
  end
end

local function damage_shield(damage, ship, shield)


  local max_shields = ship.max_shields

  if(shield=="planetary_shield_damage")
    max_shields=ship.planetary_shield
  end

  if(not ship[shield])
    return damage
  elseif(ship[shield]<max_shields) 
    damage=damage-ship["Shield Reflection"]
    if(damage<=0) 
      log("%s reflect all damage.",shield)
      return 0
    end

    local health=max_shields - ship[shield]

    if(health>damage)
      ship[shield]= ship[shield]+damage
      log("%s take %d damage (%d reflected)", 
        shield, damage, ship["Shield Reflection"])
      return 0
    else
      ship[shield]=max_shields
      damage=damage-health
      log("%s are down. (%d damage absorbed, %d reflected)", 
        shield, health, ship["Shield Reflection"])
    end

  end

  return damage

end

function max_system_and_weapon_health(ship)
  local r = 0
  for s in systems(ship) do
    --sprint(s.name,s.health,type(s.health))
    r=r + s.health
  end
  for w in weapons(ship) do
    r=r+w.health_per*w.n
  end
  return r
end

function current_system_and_weapon_health(ship,skip_munitions)
  local r = 0
  for s in systems(ship) do
    if( not ( skip_munitions and MUNITION[s.parent_type] ) )

      if not s.damage
        s.damage=0
        add_exception('missing system damage value for'..tostring(s))
      end

      r=r + s.health - s.damage
    end
  end
  for w in weapons(ship) do

     if not w.damage
        w.damage=0
        add_exception('missing system damage value for'..tostring(w))
     end

		 local _TRACE_INFO = w.parent_type


    r=r+w.health_per*w.n-w.damage
  end
  return r
end

local function ensure_critical_damage_vars(ship)

  if(ship.taken_critical_damage and ship.critical_damage)
    return
  end

  ship.taken_critical_damage=0

  ship.critical_damage = ship.structure * (pow(16,rand_01())-1)/30

end



function damage_systems(damage,ship)
  
  ensure_critical_damage_vars(ship)

  local emp_damage in FLAG

  local rval

  local health=0

  if FLAG.cripple and not emp_damage and ship.critical_condition
    log0("can't do further crippling damage to %s",ship.name)
    return 
  end

  for s in systems(ship) do

    --if( emp_damage and MUNITION[s.name] ) goto next end

    if ( s.damage < s.health)  
      health=s.health+health
    end

    ::next::

  end	

  for w in weapons(ship) do
    local whealth=w.health_per*w.n

    local partial_damage=math.fmod( w.damage, w.health_per)

    health=health + whealth - (w.damage - partial_damage)
  end

  -- if everything's already destroyed, we're done.
  if health==0 return rval end



  for s in systems(ship) do

    if(ship.critical_condition and not emp_damage and MUNITION[s.parent_type] and not FLAG.cripple)
      rval=s.parent_type
    end

    if ( s.damage < s.health ) 


      local hits = bernoulli_trials(damage,s.health/health)


      -- debuggin
      --if(SYSTEMS[s.name].shields) 
      --	hits=s.health
      --end




      if(hits+s.damage>=s.health)

        if(emp_damage)
          s.emp_damage = s.emp_damage or 0
          s.emp_damage += s.health - s.damage
          log0("%s disabled by EMP",s.parent_type)
        else
          log0("%s destroyed",s.parent_type)
        end


        s.damage=s.health


        if MUNITION[s.parent_type] and not emp_damage
            --when munitions are destroyed, they have a chance to blow up the ship
            if random_chance(.5)
              rval = s.parent_type
            end
        end

      elseif(hits>0)

        s.damage+=hits
        if(emp_damage)
          s.emp_damage = s.emp_damage or 0
          s.emp_damage += hits 
        end

        log("%s takes %d damage",s.parent_type,hits)

      end

    end

    ::next::

  end

  update_ship_speed(ship)
  update_ship_shields(ship)

  -- so, from a design perspective, it's quite important to me
  -- that the way in which weapons are packed into weapon arrays
  -- not impact their odds of being destroyed.
  --
  -- at the same time, i don't like the idea of tracking
  -- partial damage per weapon.  not only would per-weapon data
  -- badly bloat all my state serializations (which are
  -- currently concise enough to be human readable), that data 
  -- would also be awkward to visualize in the HUD, as ships
  -- commonly have so many individual weapons that a
  -- privateer-style readout would be impractical.
  --
  -- so, my solution has been to force the damage dealt to a
  -- weapon to be an exact multiple of the health per weapon,
  -- rounding up or down with probability equal to the fractional
  -- weapon damage implied by from the weapon's bernoulli_trials.
  --
  -- i calculate whether or not each weapon in the array is
  -- destroyed individually, then assign damage to the entire
  -- array based on the number of destroyed weapons.  thus, the
  -- number of weapons destroyed is clearly not influenced by the
  -- way weapons are packed into arrays.  in theory, there's
  -- probably a more efficient way to calculate the weapons
  -- destroyed per array, but, as much fun as it would be to
  -- derive an appropriate probability function, there's not much
  -- to be gained by it.
  --
  -- more than that, i'm still not entirely sold on this way of
  -- assigning weapon damage.  
  --
  for w in weapons(ship) do
    local whealth=w.health_per*w.n
    local hp = w.health_per;

    local active_weapons = (whealth - w.damage)/hp
    if(active_weapons~=ceil(active_weapons))
      error("a weapon has damage that's not an integer multiple of it's hitpoints.  this should be impossible.")
    end

    local destroyed=0
    for i=1,active_weapons,1 do
      local hits = bernoulli_trials(damage, hp/ health )
      local d,fr= modf(hits/hp)
      if(d>=1)
        destroyed=destroyed+1
      elseif(fr>0)
        if(random_chance(fr))
          destroyed=destroyed+1
        end
      end
    end

    if(destroyed>0)
      w.damage+=destroyed*hp
      if(emp_damage)
        w.emp_damage=w.emp_damage or 0
        w.emp_damage+=destroyed*hp
      end

      if(w.damage>whealth)
        error("i've done something wrong in the weapon damage logic")
      end
      log("%d x %s (%d) destroyed",destroyed,w.parent_type,hp)
    end
  end
 
  return rval

end

function deal_structure_and_system_damage(damage,ship)


  local max_health = ship.structure
  local structure = max_health-ship.structure_damage
  
  local systems_damage=damage


  if(damage>structure) 
    damage=structure
  end



  log("%d structure damage (%d)",damage,structure)

  -- check to see whether the implied systems damage causes the
  -- ship to explode.
  munitions_death = damage_systems(systems_damage,ship)



  if(structure<=0) 
    ship.structure_damage=max_health

    return  munitions_death or "Structural integrity" 
  else
    ship.structure_damage=ship.structure_damage+damage
    structure = max_health-ship.structure_damage
    
    if( get_hull(ship) < min_hull_integrity(ship) )
      ship.structure_damage=min(ship.structure_damage,max_health)
      return munitions_death or "Structural integrity" 
    end 

  end

  return munitions_death

end

local function damage_crew(damage,ship)


  if(not ship.max_crew) return end

  local hits = ceil( 3*ship.max_crew * (damage/ship.structure))

  local killed = bernoulli_trials(hits, .2)

  --sprint( damage,max_killed, frac,killed)

  local crew=ship.max_crew-ship.crew_damage

  if(crew<=0) 
    return 
  end

  if(killed<=0) 
    return 
  end

  if(crew<=killed)
    log("all remaining crew killed (%d).",crew)
    ship.crew_damage=ship.max_crew
  else
    ship.crew_damage=ship.crew_damage+killed
    log("%d crew killed.",killed)
  end

end


local function deal_armor_damage(damage,ship)

  local max_health=ship.structure
  local armor = ship.armor-ship.armor_damage

  if(armor<=0) 
    return damage 
  end


  local hit_chance=max(.5,armor/max_health)
  local reflect=ship["Armor Reflection"]
  armor_damage = min(armor+reflect,bernoulli_trials(damage, hit_chance) )
  
  local rval=damage-armor_damage

  if(armor_damage==0) return damage end

  armor_damage=max(0,armor_damage-reflect)

  if(armor_damage>0)
    ship.armor_damage=ship.armor_damage+armor_damage
    log("Armor takes %d damage. (%d reflected)",armor_damage,reflect)
    
  else
    log("The armor reflected all %d damage.",armor_damage)
  end

  return rval

end

function ShipProperty.resists_shield_piercing(ship)
  --print('calling a ship property',ship)
  local max_resistance = 0
  for s in systems(ship)
    if (s.health > s.damage) and s.resists_shield_piercing
      max_resistance=max(max_resistance, s.resists_shield_piercing)
    end
  end
  return max_resistance
end

function deal_damage(weapon, damage, ship, shield)

  -- shield damage logic

  --print('hrm', ..ship.resists_shield_piercing)

  local shield_damage_p = 1.0-
    saturate(weapon.shield_piercing_level -  ship.resists_shield_piercing)

  if shield_damage_p > 0

    local shield_damage = round | damage*shield_damage_p
    local direct_damage = damage - shield_damage

		local pre_damage = shield_damage

    if (ship.max_shields > 0) and ship[shield]
    
      if( weapon:special "Enveloping" )
        local d0 = damage_shield(shield_damage,ship,SHIELD_NAMES[1])
        local d1 = damage_shield(shield_damage,ship,SHIELD_NAMES[2])
        d0=max(d0,d1)
        local d1 = damage_shield(shield_damage,ship,SHIELD_NAMES[3])
        d0=max(d0,d1)
        local d1 = damage_shield(shield_damage,ship,SHIELD_NAMES[4])
        shield_damage=max(d0,d1)
      else
        shield_damage=damage_shield(shield_damage,ship,shield)
      end
  

    elseif(ship.planetary_shield)

      ship.planetary_shield_damage = ship.planetary_shield_damage or 0

      shield_damage=damage_shield(shield_damage,ship,"planetary_shield_damage")
    end

    -- recombine the direct and shield damage levels
    damage = direct_damage + shield_damage

		-- update capacitor_charge if needed
		if ship.capacitor_charging
			--print('adding charge',pre_damage-shield_damage)
			ship.capacitor_charge = min(ship.max_shields, pre_damage-shield_damage + (ship.capacitor_charge or 0) )
		end

  end


  local kc_damage = damage


  if(damage<=0) return end

  if weapon:special "Shield Damage Only"
    log("target's shields are gone, %d shield only damage has no effect",damage)
    return
  end

  -- crippling damage isn't prevented by armor
  if weapon:special "EMP"		
    FLAG_call(
      'emp_damage', FLAG_call, 'only_non_munitions',
      damage_systems, damage,ship)		
  end

  if(FLAG.cripple and ship.critical_condition)
    log("can't do further crippling structure damage to.")
    return
  end

  if ship.armor and not weapon:special "Armor Piercing" 
    damage=deal_armor_damage(damage,ship)
  end

  if(not ship.structure)
    error("ships must have a structure field")
  end

  if(not ship.structure_damage)
    ship.structure_damage=0
  end

  if weapon:special "Kills Crew"
    damage_crew(2*kc_damage*((ship.armor_damage or 0)/ship.armor),ship)
  end

  if(damage<=0) return end


  if(FLAG.cripple and not ship.planet)

    local safe_structure = min_hull_integrity(ship)
    if(  ship.structure - (ship.structure_damage + damage) < safe_structure) 
      damage = ship.structure - ship.structure_damage - safe_structure
      log0("truncating to %d damage",damage)

      if(damage <=0 ) 
        log0("can't do further damage to the structure without destroying the ship")
        return 
      end
    end
  end

  if not(weapon:special "Kills Crew")
    damage_crew(damage,ship)
  end

  
  local destroyed=deal_structure_and_system_damage(damage,ship)


  if(destroyed)
    log0("%s destroyed. (%s)", ship.name, MUNITION[destroyed] and MUNITION[destroyed].death_msg or destroyed)
    return destroyed
  end


  
end

function damage_roll(weapon,dist)

  local dfun = weapon.damage_fun

  if(dfun)

    local d=dfun()

    -- range attenuation for beam weapons
    if weapon.is_beam

      local range = weapon:range(dist)

      local attenuate= beam_attenuate(range)

      local old_d=d

      d=d*attenuate
      --conver
      d=floor(d+.5)

      log0("deals %d damage. (%d%% attenuation).",
        d,100-100*attenuate)
    else
      log0("deals %d damage.",d)
    end

    return d + (weapon.overcharge_damage or 0)

  end

end




local enveloping = {Enveloping=true}
local none = {}




local function doing_aoe_destruction_damage(object)
  return object.deal_destruction_aoe
--  return MUNITION[object.killed_by] or object.killed_by=="reactor overload"
end

function aoe_dist_to_threat(dist)

  if dist>3
    return 
  end

  if dist <= -0.0001
    return 1
  end

  if dist < 0.6
    return 2
  end

  if dist < 1.7
    return 3
  end

--	return 3

--	return dist<=4 and (floor( max(dist,2) )-1)
end

local threat_to_mult = {1,.6,.3}

local null_weapon = EmulateRecorded.weapons { special=_SAFE }
local enveloping_weapon = EmulateRecorded.weapons { special=function(w,k) return k=='Enveloping' end }

insert(_G,..enveloping_weapon,..null_weapon)

local function ship_aoe_damage(object)

  local to_target in range_fun
  local x,y=object_center(object)

  if not doing_aoe_destruction_damage(object)
    return
  end

  for obj in objects() do

    local dist = to_target(object,obj)

    --fconsole("%g to %s",dist,obj.name)

    local threat = aoe_dist_to_threat(dist)

    if threat
      
      --[[
      local mult=.3

      if(dist<0)
        mult=1.5
      elseif(dist<2)
        mult =1
      elseif(dist<3)
        mult=.6
      end
      --]]

      local mult = threat_to_mult![threat]*.6

      local damage = (object.hazard or 100)*mult

      log0("destruction aoe from %s hits %s for %d (dist %g, mult %g)", object.name, obj.name, damage, dist, mult)

      local starting_destroyed=obj.destroyed

      local state0=tactical_state_ind()

      local shield=facing_shield(obj,x,y)
      apply_aoe_damage(mult>1 and enveloping_weapon or null_weapon, damage,obj,shield)

    end
  end

end

function ShipProperty.hull_class(ship)
  local hull_class in ship.drawer_table
  return hull_class
end

function on_ship_death(object)


  local destruct=drawers.standard_destruction
  local current_time = current_event_time()

  if doing_aoe_destruction_damage(object)
    
    animated_object_notification {
      t0=current_event_time(), 
      tn=.7, ..object, 
      text='Reactor Detonated',text_type='critical_info',
    }

  end

  delayed_event(destruct.deathtime, function () 

    --console(..destruct,..current_time)

    advance_animation_time( current_event_time() )

    if doing_aoe_destruction_damage(object)
      local object_deltas, delta_ani = animate_deltas( .2, |) function()
        ship_aoe_damage(object)
      end
			
		  animate_destroyed_trackers_and_damage_text {
				death_type='aoe_death',
				object_changes = object_table_form(object_deltas),
				..delta_ani,
				death_delay=.5,
			}

    end

    clear_death_position(object)
    local explosion_sound = 'SHIP_EXPLOSION_MED'
    if object.hull_class >=5
      explosion_sound =  'SHIP_EXPLOSION_LRG'
    end 

    local on_scheduled = function(ani_t)
      add_triggered_sound(explosion_sound,ani_t,ani_t+.2)
    end
    local destruct_ani=inherit(destruct,
          {
            dying=object,
            ..on_scheduled,
          }
          
          )

    add_death_animation(current_time, destruct_ani)

  end)


end



local function apply_weapon_damage_to_ship(weapon, ship, shield, death_delay)


  --print(..weapon.type,..weapon.parent_type,..weapon.weapon_type,..weapon.name)
  

  -- in the case of a missile or fighter, inherit any specials
  -- set in the source objects weapon mount.
  if(is_tracking_object(weapon) and weapon.launcher_i)
    --sprint("launcher_i",weapon.launcher_i)
    weapon = weapon.src.weapons[weapon.launcher_i]
    --sprint("tracker_specials",specials)
  end

  local dfun = weapon.damage_fun

  local to_target in range_fun

  if(dfun)
    
    local wobject=weapon.object or weapon 

    if(not wobject)
      errorf("no weapon object defined for %s", tostring(weapon))
    end

    if(not is_object(wobject) )
      errorf("strange object type for weapon %s, %s", tostring(weapon), tostring(wobject) )
    end


    local d = damage_roll(weapon,to_target(wobject,ship))

    --if true return end

    local destruct=deal_damage(weapon, d, ship, shield) 

    if( destruct )
      ship.killed_by=destruct
      
      -- limit the cases in which we may do aoe damage
      ship.deal_destruction_aoe = 
        (MUNITION[ship.killed_by] or ship.killed_by=="reactor overload") 
        and true
       -- and random_chance(.3)

      if( ship.is_planet)
        ship.defenseless=true
      else
        ship.dead=true
        delayed_event(death_delay, function() on_ship_death(ship)end)
      end

      return destruct
    end
  else 
    
  end
  


end



local is_tracking_object in _G

function apply_weapon_damage(w, object, shield, death_delay)

  death_delay = death_delay or 0

  if(is_ship(object))
    return apply_weapon_damage_to_ship( w,object,shield,death_delay  )
  elseif(is_tracking_object(object))
    return apply_weapon_damage_to_tracker(w,object)
  else
    errorf("don't know how to damage objects of type %s",object.type)
  end

end


function apply_aoe_damage(weapon, damage, object, shield, death_delay)

  -- maybe planets should be immune to aoe damage?
  if(object.planet)
    return
  end


  if(object.dead)
    Log("warning, %s is already dead",object)
    return
  end

  death_delay = death_delay or 0

  if(is_ship(object))
    
    local destroyed=deal_damage(weapon, damage, object, shield)
    if(destroyed)

      if(object.planet)

        object.defenseless=true
        object.structure_damage = object.structure
        object.armor_damage = object.armor
      else

        object.killed_by=destroyed
        object.dead=true
        on_ship_death(object)

  

      end
    end

  elseif(is_tracking_object(object))
    return apply_aoe_damage_to_tracker(damage,object)
  else
    errorf("don't know how to damage objects of type %s",target.type)
  end

end